#!/usr/bin/python

import sys
import math
import os

def bipartiteMatch(graph):
        '''Find maximum cardinality matching of a bipartite graph (U,V,E).
        The input format is a dictionary mapping members of U to a list
        of their neighbors in V.  The output is a triple (M,A,B) where M is a
        dictionary mapping members of V to their matches in U, A is the part
        of the maximum independent set in U, and B is the part of the MIS in V.
        The same object may occur in both U and V, and is treated as two
        distinct vertices if this happens.'''

        # initialize greedy matching (redundant, but faster than full search)
        matching = {}
        for u in graph:
                for v in graph[u]:
                        if v not in matching:
                                matching[v] = u
                                break

        while 1:
                # structure residual graph into layers
                # pred[u] gives the neighbor in the previous layer for u in U
                # preds[v] gives a list of neighbors in the previous layer for v in V
                # unmatched gives a list of unmatched vertices in final layer of V,
                # and is also used as a flag value for pred[u] when u is in the first layer
                preds = {}
                unmatched = []
                pred = dict([(u,unmatched) for u in graph])
                for v in matching:
                        del pred[matching[v]]
                layer = list(pred)
                # repeatedly extend layering structure by another pair of layers
                while layer and not unmatched:
                        newLayer = {}
                        for u in layer:
                                for v in graph[u]:
                                        if v not in preds:
                                                newLayer.setdefault(v,[]).append(u)
                        layer = []
                        for v in newLayer:
                                preds[v] = newLayer[v]
                                if v in matching:
                                        layer.append(matching[v])
                                        pred[matching[v]] = v
                                else:
                                        unmatched.append(v)
                # did we finish layering without finding any alternating paths?
                if not unmatched:
                        unlayered = {}
                        for u in graph:
                                for v in graph[u]:
                                        if v not in preds:
                                                unlayered[v] = None
                        #return (matching,list(pred),list(unlayered))
			return matching
                # recursively search backward through layers to find alternating paths
                # recursion returns true if found path, false otherwise
                def recurse(v):
                        if v in preds:
                                L = preds[v]
                                del preds[v]
                                for u in L:
                                        if u in pred:
                                                pu = pred[u]
                                                del pred[u]
                                                if pu is unmatched or recurse(pu):
                                                        matching[v] = u
                                                        return 1
                        return 0
                for v in unmatched: recurse(v)


def treat(filename):
	"""
	Read the data file like this:
	#data_index	proc_index	color
	"""
	f = open(filename)
	mapping_ini = {}
	colors = {}	
    	for line in f.readlines():
		words = line.split()
		if words == [] :
			continue
		if words[0][0] == '#':
			continue
		data_index = words[0]
		proc_index = words[1]
		col_index = words[2]
		if proc_index not in mapping_ini:
			mapping_ini[proc_index] = []
		mapping_ini[proc_index].append(data_index)
		colors[data_index] = col_index
	return (mapping_ini, colors)


def init_bipartitegraph(mapping, colors):
	"""
	Return the complete bipartite graph with processors in one set, colors in the second set,
	and label the edges with max(to_send, to_receive) if we assigne the proc to the color
	"""
	bipartite = {}
	color_set = []
	for col in colors.values():
		if col not in color_set:
			color_set.append(col)
	for proc in mapping:
		bipartite[proc] = list(color_set)
	edges = {}
	for proc in bipartite:
		for col in bipartite[proc]:
			to_send = 0
			for data in mapping[proc]:
				if colors[data] != col:
					to_send += 1
			to_receive = 0
			for data in colors:
				if data not in mapping[proc] and colors[data] == col:
					to_receive += 1
			edges[(proc,col)] = max(to_send, to_receive)
	return (bipartite, edges)		

def marchal_herrmann(graph, edges):
	"""
	Find the matching color -> processor which minimizes the redistribution time
	"""
	nb_proc = len(graph.keys())
	for proc in graph:
		nb_color = len(graph[proc])
		break
	previous_matching = bipartiteMatch(graph)
	new_matching = previous_matching
	while len(new_matching) >= min(nb_color, nb_proc):
		previous_matching = new_matching
		max_edge = -1
		for (proc, col) in edges:
			if max_edge < edges[(proc, col)]:
				(proc_max, col_max) = (proc, col)
				max_edge = edges[(proc, col)]
		graph[proc_max].remove(col_max)
		#print "We remove ",(proc_max,col_max), " of size ", edges[(proc_max, col_max)]
		#print "graph: ",graph
		del edges[(proc_max, col_max)]
		new_matching = bipartiteMatch(graph)
		#print "New matching: ",new_matching
	return previous_matching


def print_new_file(colors, assign_col_to_proc):
	"""
	Print the file decribing the new processor index for each elements like this:
	#data_index	new_proc_index	color
	"""
	f = open(filename+".out", 'w')
	f.write("#data_index\tnew_proc_index\tcolor\n")
	for task_index in colors:
		f.write(task_index+"\t"+assign_col_to_proc[colors[task_index]]+"\t"+colors[task_index]+"\n")

def optimal_2DBC_for_parsec(filename, P, Q):
	#Take a data distribution file (formated as in Parsec) and give the optimal data distribution file for a 2DBC distribution (as in LU/QR)
	f = open(filename, 'r')
	ftemp = open(filename+".temp", 'w')
	fout = open(filename+".opt2D", 'w')
	lines = f.readlines()
	n = int(math.sqrt(len(lines)))
	for line in lines:
		words = line.split()
		i = int(words[0]) / n
		j = int(words[0]) - i*n
		ftemp.write("%d %d %d\n"%(int(words[0]), int(words[1]), ((i%P)*P + (j%Q))))
	(mapping_initial, colors) = treat(filename)
    	(bipartite_graph, edges) = init_bipartitegraph(mapping_initial, colors)
	assign_col_to_proc = marchal_herrmann(bipartite_graph, edges)
	for task_index in colors:
		fout.write(task_index+" "+assign_col_to_proc[colors[task_index]]+" "+"0\n")

def optimal_1D_for_parsec(filename, P):
	#Take a data distribution file (formated as in Parsec) and give the optimal data distribution file for a 1D distribution (as in Stencil1D)
	f = open(filename, 'r')
	ftemp = open(filename+".tmp", 'w')
	foutt = open(filename+".1D", 'w')
	fout = open(filename+".opt1D", 'w')
	lines = f.readlines()
	n = len(lines)
	nb = []
	for i in range(P-(n-P*(n/P))+1, P+1):
		nb.append(n/P+1)
	for i in range(P-(n-P*(n/P))):
		nb.append(n/P)
	color = []
	index_col = 0
	for i in nb:
		for j in range(i):
			color.append(index_col)
		index_col += 1
	print color
	for line in lines:
		words = line.split()
		ftemp.write("%d %d %d\n"%(int(words[0]), int(words[1]), color[int(words[0])-1]))
		foutt.write("%d %d %d\n"%(int(words[0]), color[int(words[0])-1], 0))
	ftemp.close()
	(mapping_initial, colors) = treat(filename+".tmp")
    	(bipartite_graph, edges) = init_bipartitegraph(mapping_initial, colors)
	assign_col_to_proc = marchal_herrmann(bipartite_graph, edges)
	for task_index in sorted(colors, key=int):
		fout.write(task_index+" "+assign_col_to_proc[colors[task_index]]+" "+"0\n")
	f.close()
	fout.close()
	foutt.close()
	os.remove(filename+".tmp")
		

if __name__ == '__main__':
	"""
    	filename = sys.argv[1]
	(mapping_initial, colors) = treat(filename)
    	(bipartite_graph, edges) = init_bipartitegraph(mapping_initial, colors)
	assign_col_to_proc = marchal_herrmann(bipartite_graph, edges)
	print_new_file(colors, assign_col_to_proc)
	print "couleur -> proc :",assign_col_to_proc
	"""
	for i in range(10):
		optimal_1D_for_parsec("../../../../rand_distrib/distrib_%d"%(i),16)


